#include <ctype.h>
#include <limits.h>
#include <string.h>
#include <sys/stat.h>
#include <syslog.h>
#include <arpa/inet.h>  // needed for in_addr struct
#include <time.h>    // needed for radutmp.h
#include "radutmp.h"
#include "common.h"

#define   MSG_DEFAULT           0  // Unknown error.
#define   MSG_SESSION           1  // No time left this session.
#define   MSG_TIME_LEFT         2  // No time left.
#define   MSG_TIME_CLASS        3  // No time left in current time class.
#define   MSG_LOCK              4  // User is locked out.
#define   MSG_NO_RECORD         5  // No ACUA record for this user.
#define   MSG_NO_ACCOUNT        6  // No login in systems /etc/passwd.
#define   MSG_BAD_ARGC          7  // Incorrect usage of acua_radlogin.
#define   MSG_HOST              8  // Couldn't look up our own hostname.
#define   MSG_LONG_NAME         9  // Users login name is too long.
#define   MSG_UNSUBSCRIBED      10 // User has been unsubscribed.
#define   MSG_TOOMANY           11 // Users maxLogins would be exceeded.
#define   MSG_BYTE_LIMIT        12 // User has exceeded byte limit.
#define   MSG_NO_SUBSCRIPTION   13 // User does not have a subscription.

/* This is the amount of time given to users which
 * do not have any time left, IN SECONDS. A setting of 0 will mean that the
 * login will be refused.
 */
#define   NoTimeLeft            0

/*[ End User Defines ]******************************************************/

void    accept();
void    reject();

// Global Variables.

int                      nLoginRecs,
                         nProcs,
                         msg = -1,
                         curClass = -1,
                         i,
                         sLeft;
LoginRec                 loginRec[MAX_USERS];
ProcRec                  procRec[MAX_PROCS];
char                     RejectReason[100],
                         hostname[256],
                         login[MAX_LOGINCHARS + 1];
UserRec                  ur;
Boolean                  userRecFound = FALSE;
uid_t                    uid;
HashTable                loginRecTable;
word                     myHostAddr;
int                      hostIdx;
struct     radutmp       rt;
struct     in_addr       __in_addr;
/***************************************************************************/

int
main(int argc, char **argv)
{

  // Open the syslog.
  openlog("acua_radlogin", LOG_PID, LOG_AUTHPRIV);

  /* Check usage is correct */
  if (argc < 2) {
    printf("Usage: acua_radlogin <login>\n");
    // Send the error info to the log also.
    exit(1);
  }

  gethostname(hostname, 256);
  myHostAddr = hostAddr(hostname);

  readConfig();
  userFileOpen();
  userList(&nLoginRecs, loginRec);
  procList(&nProcs, procRec);

  // Check the length of argv[1]
  if (strlen(argv[1])>MAX_LOGINCHARS) { msg = MSG_LONG_NAME;
                                        reject();
                                      }

  // Set username so that the rest of the application can see it.
  strncpy(login, argv[1], MAX_LOGINCHARS);
  //username[MAX_LOGINCHARS]='\0';

  // Get the uid of the user we're logging in.
  uid = UIDfromLogin(login);

  // Check to make sure this user exists.
  if (uid<0) {
    msg = MSG_NO_ACCOUNT;
    reject();
  }

  // find their user record
  if (userFileSearch(&ur, uid)) { msg = MSG_NO_RECORD;
                                  reject();
  }
  // user is found
  userRecFound = TRUE;
  if (EXPIRE(ur.flags) == EXPIRE_DELETE) { msg=MSG_UNSUBSCRIBED;
                                           reject();
  }
  
  // reject if they do not have a subscription
  if (!ur.expire) { msg = MSG_NO_SUBSCRIPTION;
                    reject();
  }
  // If they are already logged in then check to see if they are online
  // according to radius's utmp file.
  if ((ur.nHosts >= ur.maxLogins) && (radiusTotalLogins(login) > ur.maxLogins))
  {
      msg=MSG_TOOMANY;
      reject();
  }
  // are they locked out?
  if (LOCK(ur.flags)) { msg = MSG_LOCK;
                        reject();
  }
  // are they over their byte limit?
  if ((ur.bLimit && ur.bTx + ur.bRx > ur.bLimit) ||
      (ur.bTxLimit && ur.bTx > ur.bTxLimit) ||
      (ur.bRxLimit && ur.bRx > ur.bRxLimit)) {
    msg = MSG_BYTE_LIMIT;
    reject();
  }
  // give them an ip?
  if (ur.flagz & FLG_NOIP)
    accept();
  // are they a Multilink user?
  if (ur.flags & FLG_MULTILINK) {
    if (radiusCheckOnline(&rt, login))
      {
      __in_addr.s_addr = rt.framed_address;
      printf("Framed-IP-Address = %s\n", inet_ntoa(__in_addr));
      } else {
      printf("Framed-IP-Address = 255.255.255.254\n");
      }
  } else printf("Framed-IP-Address = 255.255.255.254\n");
  // if all above tests passed then it should be ok to accept.
  accept();
}

/***************************************************************************/

void reject()
{

  // Figure out the possible reasons for why the user was rejected
  // if the reason hasn't been set already.
  if (msg < 0) {
    if (ur.tLeft <= 0)
      msg = MSG_TIME_LEFT;
    else if (curClass >= 0 && ur.cLimit[curClass] >= 0 && ur.cLeft[curClass] <= 0)
      msg = MSG_TIME_CLASS;
    else
      msg = MSG_DEFAULT;
  }

  // determine if the user is being rejected due to time problems.
  if ((msg==MSG_TIME_LEFT) || (msg==MSG_TIME_CLASS) || (msg==MSG_SESSION)) {
    if (NoTimeLeft==0) {
      userFileClose();
      exit(1);
    }
    ur.tLeft = NoTimeLeft;
    ur.sLeft = NoTimeLeft;    // accept() will write these new values.
    accept();
  }

  // close the userbase since it's not needed anymore.
  userFileClose();

  // Figure out login rejection reason.
  // RejectReason, by default, is only 100 chars long, so be careful.
  if (msg==MSG_DEFAULT) sprintf(RejectReason,      "Unknown, possible error.");
  if (msg==MSG_LOCK) sprintf(RejectReason,         "Account locked.");
  if (msg==MSG_TIME_LEFT) sprintf(RejectReason,    "Time limit exceeded.");
  if (msg==MSG_TIME_CLASS) sprintf(RejectReason,   "No time left in current class.");
  if (msg==MSG_SESSION) sprintf(RejectReason,      "Session time limit exceeded.");
  if (msg==MSG_NO_RECORD) sprintf(RejectReason,    "User record not found in ACUA user file.");
  if (msg==MSG_NO_ACCOUNT) sprintf(RejectReason,   "User account not found in /etc/passwd file.");
  if (msg==MSG_BAD_ARGC) sprintf(RejectReason,     "Incorrect usage of acua_radlogin!");
  if (msg==MSG_HOST) sprintf(RejectReason,         "Could not look up our own hostname.");
  if (msg==MSG_LONG_NAME) sprintf(RejectReason,    "Users login name is too long.");
  if (msg==MSG_UNSUBSCRIBED) sprintf(RejectReason, "User has been unsubscribed.");
  if (msg==MSG_TOOMANY) sprintf(RejectReason,      "User has exceeded maximum logins.");
  if (msg==MSG_BYTE_LIMIT) sprintf(RejectReason,   "User has exceeded maximum byte limit.");
  if (msg==MSG_NO_SUBSCRIPTION) sprintf(RejectReason,   "User does not have a subscription.");

  // log a message stating why the users login was rejected
  if (msg != MSG_BAD_ARGC)
    syslog(LOG_NOTICE, "User [%s] Login Rejected: %s", login, RejectReason);
  else
    syslog(LOG_NOTICE, "Login Rejected: %s", RejectReason);
  closelog();
  exit(1);
}

/***************************************************************************/

void accept()
{

  if (!userRecFound) {
    msg = MSG_NO_RECORD;
    reject();
  }

  // Get our hostname.
  if (gethostname(hostname, 256)) {
    msg = MSG_HOST;
    reject();
  }

  hashTableInit(&loginRecTable, MAX_USERS, UIDhash, UIDcomp, 0, FALSE);
  for (i = 0; i < nLoginRecs; i++)
    hashTableAdd(&loginRecTable, &loginRec[i]);

  addHost(&ur, myHostAddr);
  hostIdx = findHost(&ur, myHostAddr);
  ur.lastLogin[hostIdx] = time(NULL);

  // Update the last login info.
  // this should prolly be done when marking offline.
  ur.lastLogin[hostIdx] = time(NULL);
  ur.lastOnline[hostIdx] = time(NULL);

  // Work out how long the user is allowed to be on-line for.
  sLeft = INT_MAX;
  if (ur.tLimit >= 0) sLeft = min(sLeft, ur.tLeft + ur.credit);
  if (ur.sLimit >= 0) sLeft = min(sLeft, ur.sLimit);
  if (curClass >= 0 && ur.cLimit[curClass] >= 0) sLeft = min(sLeft, ur.cLeft[curClass]);

  // If the user has no time left, then disallow the login.
  if (sLeft <= 0) {
    msg = MSG_SESSION;
    reject();           // FIXME:  Maybe we could get a loop here cause we
  }                     // are calling reject(), but reject() can call us!

  // Set up some more connection info.
  if (ur.PPPidleMinutes > 0)
  {
  
    if (ur.PPPidleMinutes < 120)
        printf("Idle-Timeout = 120\n");
    
    if (ur.PPPidleMinutes * 60 > 14400)
        printf("Idle-Timeout = 14400\n");
    
    if ((ur.PPPidleMinutes >= 120) &&
	(ur.PPPidleMinutes * 60 <= 14400))
	    printf("Idle-Timeout = %d\n", ur.PPPidleMinutes * 60);
  }
  

  syslog(LOG_INFO, "User [%s] logged in with %d minutes left", login, sLeft);
  ur.sLeft = ur.sLimit;
  ur.bStx = ur.bSrx = 0;
  userFileEdit(&ur);
  userFileClose();
  hashTableDeinit(&loginRecTable);
  closelog();
  exit(0);
}

